/***
* ==++==
*
* 
*
* ==--==
* =+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+
*
* oAuth.h
*
* This file contains helper classes for oAuth authorization and authentication.
* 
*
* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
****/
#ifndef OAUTH_H
#define OAUTH_H
#endif

#include <stdint.h>
#include <limits>
#include <string>
#include <vector>
#include <memory>
#include <random>
#include <ctime>
#include <chrono>
#include <array>

#include "uri.h"
#include "asyncrt_utils.h"
#include "http_client.h"


#define SIZEOF(x) (sizeof(x)/sizeof(*x))

namespace Authentication
{	

    // Dropbox consumer key and secret
    static std::wstring consumerKey = L"Your app key";
    static std::wstring consumerSecret = L"Your app secret";

    // List of Dropbox Authenticate, Authorize, Request and File Upload URIs
    const std::wstring DropBoxRequestTokenURI = 
		L"https://api.dropbox.com/1/oauth/request_token";
    const std::wstring DropBoxAuthorizeURI = 
		L"https://www.dropbox.com/1/oauth/authorize";
    const std::wstring DropBoxAccessTokenURI = 
		L"https://api.dropbox.com/1/oauth/access_token";
    const std::wstring DropBoxFileUploadURI = 
		L"https://api-content.dropbox.com/1/files_put/dropbox/sample.txt";
    static std::wstring LocalFiletoUpload = L"Your local file goes here";

    typedef std::map<std::wstring, std::wstring> HTTPParameters;
  	
    const std::wstring unreservedChars = L"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-_.~";

	static void Split(const std::wstring& str, std::vector<std::wstring>& out, TCHAR sep, bool includeEmpty)
	{
		unsigned start = 0;
		unsigned end = 0;

		while(true)
		{
			if(end == str.size() || str[end] == sep)
			{
				if(end > start || includeEmpty)
				{
					out.push_back(str.substr(start, end - start));
				}

				if(end == str.size())
				{
					break;
				}

				++end;
				start = end;
			}
			else
			{
				++end;
			}
		}
	}

    
    class OAuth
    {

    public:	
	    OAuth()
	    {
	    }

	    // CreateOAuthSignedParameters used for all OAuth related queries
	    //
	    // consumerKey and consumerSecret - must be provided for every call, they identify the application
	    // oauthToken and oauthTokenSecret - need to be provided for every call, except for the first token request before authorizing		
	    HTTPParameters CreateOAuthSignedParameters( 
		    const std::wstring& url, 
		    const std::wstring& httpMethod, 
		    const HTTPParameters* postParameters,
		    const std::wstring& consumerKey, 
		    const std::wstring& consumerSecret, 
		    const std::wstring& oauthToken = L"", 
		    const std::wstring& oauthTokenSecret = L""
			
		    )
	    {
		    web::http::uri u(url);
		    std::wstring query = u.query();
		    HTTPParameters getParameters = ParseQueryString(query);

		    auto signedparams =  BuildSignedOAuthParameters(
			    getParameters,
			    url,
			    httpMethod,
			    postParameters,
			    consumerKey, consumerSecret,
			    oauthToken, oauthTokenSecret
			    );

		    return signedparams; 
	    }

	    std::wstring OAuthFindHeaderValue( const std::wstring& value ) 
	    {			
		    auto it = m_oauthParameters.find(value);
		    if (it != m_oauthParameters.end())
		    {
			    return it->second;
		    }	
		    return L"";
	    }
		
	    std::wstring OAuthBuildSignedHeaders(const web::http::uri& u)
	    {
		    std::wstring query = u.to_string();
		    query += L"?oauth_consumer_key=" + consumerKey + L"&";
		    query += L"oauth_nonce=" + OAuthFindHeaderValue(L"oauth_nonce") + L"&";		
		    query += L"oauth_timestamp=" + OAuthFindHeaderValue(L"oauth_timestamp") + L"&";	
		    query += L"oauth_version=1.0&";
		    query += L"oauth_signature_method=PLAINTEXT&";
		    query += L"oauth_signature=" + OAuthFindHeaderValue(L"oauth_signature");
			
		    if (!OAuthFindHeaderValue(L"oauth_token").empty())
		    {
			    query += L"&oauth_token=" + OAuthFindHeaderValue(L"oauth_token");
		    }
		    return query;
	    }
private:

	    // create oauth requestParameters
	    HTTPParameters m_oauthParameters;
		
	    HTTPParameters ParseQueryString( const std::wstring& url ) 
	    {
		    HTTPParameters ret;

		    std::vector<std::wstring> queryParams;
		    Split(url, queryParams, L'&', false);

		    for(auto& i : queryParams)
		    {
			    std::vector<std::wstring> paramElements;
			    Split(i, paramElements, L'=', true);				
			    if(paramElements.size() == 2)
			    {
				    ret[paramElements[0]] = paramElements[1];
			    }
		    }
		    return ret;
	    }
		
        HTTPParameters BuildSignedOAuthParameters( 
		  const HTTPParameters& requestParameters,
		  const std::wstring& url,
		  const std::wstring& httpMethod,
		  const HTTPParameters* postParameters,
		  const std::wstring& consumerKey,
		  const std::wstring& consumerSecret,
		  const std::wstring& requestToken = L"",
		  const std::wstring& requestTokenSecret = L""
		  )
		{
		  std::wstring timestamp = OAuthCreateTimestamp();
		  std::wstring nonce = OAuthCreateNonce();			
		  m_oauthParameters[L"oauth_timestamp"] = timestamp;
		  m_oauthParameters[L"oauth_nonce"] = nonce;
		  m_oauthParameters[L"oauth_version"] = L"1.0";
		  m_oauthParameters[L"oauth_signature_method"] = L"PLAINTEXT";
		  m_oauthParameters[L"oauth_consumer_key"] = consumerKey;

		  // Add the request token if found
		  if (!requestToken.empty())
		  {
			m_oauthParameters[L"oauth_token"] = requestToken;
		  }			
		  // Create a parameter list containing both oauth and original parameters
		  // This will be used to create the parameter signature
		  HTTPParameters allParameters = requestParameters;
		  if(Compare(httpMethod, L"POST", false) && postParameters)
		  {
			allParameters.insert(postParameters->begin(), postParameters->end());
		  }
		  allParameters.insert(m_oauthParameters.begin(), m_oauthParameters.end());

		  // Prepare a signature base, a carefully formatted string containing 
		  // all of the necessary information needed to generate a valid signature
		  std::wstring normalUrl = OAuthNormalizeUrl(url);
		  std::wstring normalizedParameters = 
			OAuthNormalizeRequestParameters(allParameters);

		  std::wstring signatureBase = 
			OAuthConcatenateRequestElements(httpMethod, 
			normalUrl, 
			normalizedParameters);
		  // Obtain a signature and add it to header requestParameters
		  std::wstring signature = OAuthCreateSignature(signatureBase, 
			consumerSecret,
			requestTokenSecret);

		  m_oauthParameters[L"oauth_signature"] = UrlEncode(signature);
		  return m_oauthParameters;
		}


	    std::wstring OAuthNormalizeUrl( const std::wstring& url ) 
	    {
		    std::wstring normalizedUrl = L"";
		    web::http::uri u(url);
		    normalizedUrl = u.scheme() + L"://" + u.host();			
		    normalizedUrl += u.path();

		    return normalizedUrl;

	    }

	    std::wstring OAuthNormalizeRequestParameters( const HTTPParameters& requestParameters ) 
	    {
		    std::list<std::wstring> sorted;
		    for(auto& it : requestParameters)
		    {
			    std::wstring param = it.first + L"=" + it.second;
			    sorted.push_back(param);
		    }
		    sorted.sort();

		    std::wstring params;
		    for(auto it : sorted)
		    {
			    if(params.size() > 0)
			    {
				    params += L"&";
			    }
			    params += it;
		    }

		    return params;
	    }

	    std::wstring OAuthConcatenateRequestElements( const std::wstring& httpMethod, std::wstring url, const std::wstring& parameters ) 
	    {
		    std::wstring escapedUrl = UrlEncode(url);
		    std::wstring escapedParameters = UrlEncode(parameters);

		    std::wstring ret = httpMethod + L"&" + escapedUrl + L"&" + escapedParameters;
		    return ret;
	    }
			

	    std::wstring OAuthCreateSignature( const std::wstring& signatureBase, const std::wstring& consumerSecret, const std::wstring& requestTokenSecret ) 
	    {
			UNREFERENCED_PARAMETER(signatureBase);
		    // URL encode key elements
		    std::wstring escapedConsumerSecret = UrlEncode(consumerSecret);
		    std::wstring escapedTokenSecret = UrlEncode(requestTokenSecret);

		    std::wstring key = escapedConsumerSecret + L"&" + escapedTokenSecret;
		    //using PLAINTEXT signature type
		    return key;			
	    }
		
	    /// <summary>
        /// Generate a nonce
        /// </summary>
        /// <returns></returns>
        std::wstring OAuthCreateNonce() 
	    {
            wchar_t ALPHANUMERIC[] = L"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
		    std::wstring nonce;

		    for(int i = 0; i <= 16; ++i)
		    {
			    nonce += ALPHANUMERIC[rand() % (SIZEOF(ALPHANUMERIC) - 1)]; // don't count null terminator in array
		    }
		    return nonce;
        }

	    /// <summary>
        /// Generate the timestamp for the signature        
        /// </summary>
        /// <returns></returns>
        std::wstring OAuthCreateTimestamp() 
	    {
		    __time64_t utcNow;
		    __time64_t ret = _time64(&utcNow);
		    _ASSERTE(ret != -1);

		    // return the seconds as string here
		    std::wstringstream ss;
		    ss << utcNow;
		    return ss.str();
        }
			
	    std::wstring Char2Hex( wchar_t dec )
	    {
		    wchar_t dig1 = (dec&0xF0)>>4;			 
		    wchar_t dig2 = (dec&0x0F);
			 
		    if ( 0<= dig1 && dig1<= 9) dig1+=48;    //0,48 in ascii
		    if (10<= dig1 && dig1<=15) dig1+=65-10; //A,65 in ascii
		    if ( 0<= dig2 && dig2<= 9) dig2+=48;
		    if (10<= dig2 && dig2<=15) dig2+=65-10;			

		    std::wstring r;
		    r.append( &dig1, 1);
		    r.append( &dig2, 1);
		    return r;
	    }


	    // <summary>
        /// This is a Url Encode implementation.
	    /// 	
        /// </summary>
        /// <param name="value">The value to Url encode</param>
        /// <returns>Returns a Url encoded string</returns>        
        std::wstring UrlEncode(std::wstring value) {
            std::wstring result;

            for (auto symbol : value) {
                if (unreservedChars.find(symbol) != -1) {
				    result += symbol;
                } else 
			    {	
				    result.append(L"%");
				    result.append( Char2Hex(symbol) );//converts char 255 to string "ff"
                }
            }

            return result;
        }

protected:

	    inline bool Compare(const std::wstring& one, const std::wstring& two, bool caseSensitive)
	    {
		    return caseSensitive ? (one == two) : (_wcsicmp(one.c_str(), two.c_str()) == 0);
	    }		
		
};
}